@@ -18,25 +18,28 @@ from urllib.parse import urlencode
 from speechmatics.rt import AsyncClient
 from speechmatics.rt import AudioEncoding
 from speechmatics.rt import AudioFormat
+from speechmatics.rt import ClientMessageType
 from speechmatics.rt import ConversationConfig
 from speechmatics.rt import ServerMessageType
 from speechmatics.rt import TranscriptionConfig
 
 from . import __version__
+from ._audio import AudioBuffer
 from ._logging import get_logger
 from ._models import AgentClientMessageType
 from ._models import AgentServerMessageType
 from ._models import AnnotationFlags
 from ._models import AnnotationResult
+from ._models import ClientSessionInfo
 from ._models import DiarizationFocusMode
 from ._models import DiarizationSpeakerConfig
 from ._models import EndOfUtteranceMode
 from ._models import FragmentUtils
 from ._models import LanguagePackInfo
-from ._models import SessionInfo
 from ._models import SpeakerSegmentView
 from ._models import SpeakerVADStatus
 from ._models import SpeechFragment
+from ._models import TranscriptionUpdatePreset
 from ._models import VoiceAgentConfig
 
 DEBUG_MORE = os.getenv("SPEECHMATICS_DEBUG_MORE", "0").lower() in ["1", "true"]
@@ -90,13 +93,33 @@ class VoiceAgentClient(AsyncClient):
         self._is_ready_for_audio: bool = False
 
         # Session info (updated on session created)
-        self._session: SessionInfo = SessionInfo(
+        self._client_session: ClientSessionInfo = ClientSessionInfo(
             config=self._config,
             session_id="NOT_SET",
             base_time=datetime.datetime.now(datetime.timezone.utc),
-            language_pack_info=LanguagePackInfo.from_dict({}),
+            language_pack_info=LanguagePackInfo.model_validate({}),
         )
 
+        # Change filter to emit segments
+        self._change_filter: list[AnnotationFlags] = [AnnotationFlags.NEW]
+        # Full text has changed
+        if self._config.transcription_update_preset == TranscriptionUpdatePreset.COMPLETE:
+            self._change_filter.append(AnnotationFlags.UPDATED_FULL)
+        # Full text and timing have changed
+        elif self._config.transcription_update_preset == TranscriptionUpdatePreset.COMPLETE_PLUS_TIMING:
+            self._change_filter.append(AnnotationFlags.UPDATED_FULL)
+            self._change_filter.append(AnnotationFlags.UPDATED_WORD_TIMINGS)
+        # Word content only has changed
+        elif self._config.transcription_update_preset == TranscriptionUpdatePreset.WORDS:
+            self._change_filter.append(AnnotationFlags.UPDATED_STRIPPED)
+        # Word content and timing have changed
+        elif self._config.transcription_update_preset == TranscriptionUpdatePreset.WORDS_PLUS_TIMING:
+            self._change_filter.append(AnnotationFlags.UPDATED_STRIPPED)
+            self._change_filter.append(AnnotationFlags.UPDATED_WORD_TIMINGS)
+        # Timing only has changed
+        elif self._config.transcription_update_preset == TranscriptionUpdatePreset.TIMING:
+            self._change_filter.append(AnnotationFlags.UPDATED_WORD_TIMINGS)
+
         # STT message received queue
         self._stt_message_queue: asyncio.Queue[Callable[[], Awaitable[None]]] = asyncio.Queue()
         self._stt_queue_task: Optional[asyncio.Task] = None
@@ -138,11 +161,19 @@ class VoiceAgentClient(AsyncClient):
         self._metrics_emitter_task: Optional[asyncio.Task] = None
 
         # Audio sampling info
-        self._audio_sample_rate: float = self._audio_format.sample_rate * 1.0
-        self._audio_sample_width: float = {
-            AudioEncoding.PCM_F32LE: 4.0,
-            AudioEncoding.PCM_S16LE: 2.0,
-        }.get(self._audio_format.encoding, 1.0)
+        self._audio_sample_rate: int = self._audio_format.sample_rate
+        self._audio_sample_width: int = {
+            AudioEncoding.PCM_F32LE: 4,
+            AudioEncoding.PCM_S16LE: 2,
+        }.get(self._audio_format.encoding, 1)
+
+        # Audio buffer
+        if self._config.audio_buffer_length > 0:
+            self._audio_buffer: AudioBuffer = AudioBuffer(
+                sample_rate=self._audio_format.sample_rate,
+                frame_size=self._audio_format.chunk_size,
+                total_seconds=self._config.audio_buffer_length,
+            )
 
         # Register handlers
         self._register_event_handlers()
@@ -197,7 +228,9 @@ class VoiceAgentClient(AsyncClient):
             if config.prefer_current_speaker is not None:
                 dz_cfg["prefer_current_speaker"] = config.prefer_current_speaker
             if config.known_speakers:
-                dz_cfg["speakers"] = {s.label: s.speaker_identifiers for s in config.known_speakers}
+                dz_cfg["speakers"] = [
+                    {"label": s.label, "speaker_identifiers": s.speaker_identifiers} for s in config.known_speakers
+                ]
             if config.max_speakers is not None:
                 dz_cfg["max_speakers"] = config.max_speakers
             if dz_cfg:
@@ -231,20 +264,20 @@ class VoiceAgentClient(AsyncClient):
         """
 
         # Recognition started event
-        @self.once(ServerMessageType.RECOGNITION_STARTED)
+        @self.once(ServerMessageType.RECOGNITION_STARTED)  # type: ignore[misc]
         def _evt_on_recognition_started(message: dict[str, Any]) -> None:
             if DEBUG_MORE:
                 self._logger.debug(json.dumps(message))
             self._is_ready_for_audio = True
-            self._session = SessionInfo(
+            self._client_session = ClientSessionInfo(
                 config=self._config,
                 session_id=message.get("id", "UNKNOWN"),
                 base_time=datetime.datetime.now(datetime.timezone.utc),
-                language_pack_info=LanguagePackInfo.from_dict(message.get("language_pack_info", {})),
+                language_pack_info=LanguagePackInfo.model_validate(message.get("language_pack_info", {})),
             )
 
         # Partial transcript event
-        @self.on(ServerMessageType.ADD_PARTIAL_TRANSCRIPT)
+        @self.on(ServerMessageType.ADD_PARTIAL_TRANSCRIPT)  # type: ignore[misc]
         def _evt_on_partial_transcript(message: dict[str, Any]) -> None:
             if DEBUG_MORE:
                 self._logger.debug(json.dumps(message))
@@ -255,7 +288,7 @@ class VoiceAgentClient(AsyncClient):
             self._stt_message_queue.put_nowait(_handle)
 
         # Final transcript event
-        @self.on(ServerMessageType.ADD_TRANSCRIPT)
+        @self.on(ServerMessageType.ADD_TRANSCRIPT)  # type: ignore[misc]
         def _evt_on_final_transcript(message: dict[str, Any]) -> None:
             if DEBUG_MORE:
                 self._logger.debug(json.dumps(message))
@@ -266,7 +299,7 @@ class VoiceAgentClient(AsyncClient):
             self._stt_message_queue.put_nowait(_handle)
 
         # End of utterance event
-        @self.on(ServerMessageType.END_OF_UTTERANCE)
+        @self.on(ServerMessageType.END_OF_UTTERANCE)  # type: ignore[misc]
         def _evt_on_end_of_utterance(message: dict[str, Any]) -> None:
             if DEBUG_MORE:
                 self._logger.debug(json.dumps(message))
@@ -314,6 +347,11 @@ class VoiceAgentClient(AsyncClient):
         if not self._is_connected:
             return
 
+        # Send end of transcript to release resources
+        await self.send_message(
+            {"message": ClientMessageType.END_OF_STREAM, "last_seq_no": self._session.sequence_number}
+        )
+
         # Disconnect from API
         try:
             await asyncio.wait_for(self.close(), timeout=5.0)
@@ -328,9 +366,22 @@ class VoiceAgentClient(AsyncClient):
             self._is_ready_for_audio = False
             self._stop_metrics_task()
 
-        # Stop the task
+        # Stop the finalize task
+        if self._finalize_task:
+            self._finalize_task.cancel()
+            try:
+                await self._finalize_task
+            except asyncio.CancelledError:
+                pass
+            self._finalize_task = None
+
+        # Stop the STT queue task
         if self._stt_queue_task:
             self._stt_queue_task.cancel()
+            try:
+                await self._stt_queue_task
+            except asyncio.CancelledError:
+                pass
             self._stt_queue_task = None
 
     def update_diarization_config(self, config: DiarizationSpeakerConfig) -> None:
@@ -352,9 +403,9 @@ class VoiceAgentClient(AsyncClient):
     async def _run_stt_queue(self) -> None:
         """Run the STT message queue."""
         while True:
-            callback = await self._stt_message_queue.get()
-
             try:
+                callback = await self._stt_message_queue.get()
+
                 if asyncio.iscoroutine(callback):
                     await callback
                 elif asyncio.iscoroutinefunction(callback):
@@ -364,6 +415,12 @@ class VoiceAgentClient(AsyncClient):
                     if asyncio.iscoroutine(result):
                         await result
 
+            except asyncio.CancelledError:
+                self._logger.debug("STT queue task cancelled")
+                return
+            except RuntimeError:
+                self._logger.debug("STT queue event loop closed")
+                return
             except Exception:
                 self._logger.warning("Exception in STT message queue", exc_info=True)
 
@@ -389,6 +446,10 @@ class VoiceAgentClient(AsyncClient):
         # Send to the AsyncClient
         await super().send_audio(payload)
 
+        # Add to audio buffer (use put_bytes to handle variable chunk sizes)
+        if self._config.audio_buffer_length > 0:
+            await self._audio_buffer.put_bytes(payload)
+
         # Calculate the time (in seconds) for the payload
         if self._audio_format is not None:
             self._total_bytes += len(payload)
@@ -641,17 +702,15 @@ class VoiceAgentClient(AsyncClient):
 
             # Compare previous view to this view
             if self._previous_view:
-                changes = FragmentUtils.compare_views(self._session, self._previous_view, self._current_view)
+                changes = FragmentUtils.compare_views(self._client_session, self._previous_view, self._current_view)
             else:
                 changes = AnnotationResult.from_flags(AnnotationFlags.NEW)
 
             # Update the previous view
             self._previous_view = self._current_view
 
-        # Catch no changes for ADAPTIVE mode
-        if self._end_of_utterance_mode == EndOfUtteranceMode.ADAPTIVE and not changes.any(
-            AnnotationFlags.NEW, AnnotationFlags.UPDATED_FULL
-        ):
+        # Catch no changes
+        if not changes.any(*self._change_filter):
             return
 
         # Emit the segments
@@ -664,11 +723,20 @@ class VoiceAgentClient(AsyncClient):
             # Calculate delay to end of utterance
             should_emit, emit_final_delay = self._calc_delay_to_end_of_utterance(self._current_view, changes)
 
+            # Debug
+            if DEBUG_MORE:
+                self._logger.debug(
+                    f"changes: {changes}, should_emit: {should_emit}, emit_final_delay: {emit_final_delay}"
+                )
+
             # Emit finalized segments
             async def emit(delay: float | None) -> None:
-                if delay and delay > 0:
-                    await asyncio.sleep(delay)
-                asyncio.create_task(self._emit_segments(finalize=True, end_of_turn=True))
+                try:
+                    if delay and delay > 0:
+                        await asyncio.sleep(delay)
+                    await self._emit_segments(finalize=True, end_of_turn=True)
+                except asyncio.CancelledError:
+                    pass
 
             # Finalize after delay
             if should_emit:
@@ -679,7 +747,7 @@ class VoiceAgentClient(AsyncClient):
     def _update_current_view(self) -> None:
         """Load the current view of the speech fragments."""
         self._current_view = SpeakerSegmentView(
-            session=self._session,
+            session=self._client_session,
             fragments=self._speech_fragments.copy(),
             focus_speakers=self._dz_config.focus_speakers,
         )
@@ -715,7 +783,7 @@ class VoiceAgentClient(AsyncClient):
         last_active_segment = view.segments[last_active_segment_index] if last_active_segment_index > -1 else None
 
         # If this is NEW or UPDATED_FULL_LCASE
-        if view_changes.any(AnnotationFlags.NEW, AnnotationFlags.UPDATED_FULL_LCASE):
+        if view_changes.any(*self._change_filter):
             """Process the annotation flags to determine how long before sending a final segment."""
 
             # Fallback when using FIXED
@@ -787,6 +855,9 @@ class VoiceAgentClient(AsyncClient):
             ttl: Optional delay before finalizing partial segments.
         """
 
+        if DEBUG_MORE:
+            self._logger.debug("external finalise() called")
+
         async def emit() -> None:
             # Yield until the delay has passed
             if ttl is not None and ttl > 0:
@@ -818,9 +889,6 @@ class VoiceAgentClient(AsyncClient):
         # Lock the speech fragments
         if self._current_view and self._current_view.segment_count > 0:
             async with self._speech_fragments_lock:
-                # Metadata
-                metadata = {"start_time": self._current_view.start_time, "end_time": self._current_view.end_time}
-
                 # Force finalize
                 if finalize:
                     final_segments = self._current_view.segments
@@ -829,20 +897,23 @@ class VoiceAgentClient(AsyncClient):
                 # Split between finals and interim segments
                 else:
                     final_segments = [
-                        s
-                        for s in self._current_view.segments
-                        if s.annotation.has(AnnotationFlags.ENDS_WITH_FINAL, AnnotationFlags.ENDS_WITH_EOS)
+                        s for s in self._current_view.segments if s.annotation.has(AnnotationFlags.ENDS_WITH_FINAL)
                     ]
                     interim_segments = [s for s in self._current_view.segments if s not in final_segments]
 
                 # Emit finals first
                 if final_segments:
+                    # Metadata for final segments uses actual start/end times of the segments being emitted
+                    final_metadata = {
+                        "start_time": final_segments[0].start_time,
+                        "end_time": final_segments[-1].end_time,
+                    }
                     self.emit(
                         AgentServerMessageType.ADD_SEGMENT,
                         {
                             "message": AgentServerMessageType.ADD_SEGMENT.value,
-                            "segments": [s.to_dict(self._config.include_results) for s in final_segments],
-                            "metadata": metadata,
+                            "segments": [s.model_dump(self._config.include_results) for s in final_segments],
+                            "metadata": final_metadata,
                         },
                     )
                     self._trim_before_time = final_segments[-1].end_time
@@ -852,12 +923,17 @@ class VoiceAgentClient(AsyncClient):
 
                 # Emit interim segments
                 if interim_segments:
+                    # Metadata for partial segments uses actual start/end times of the segments being emitted
+                    partial_metadata = {
+                        "start_time": interim_segments[0].start_time,
+                        "end_time": interim_segments[-1].end_time,
+                    }
                     self.emit(
                         AgentServerMessageType.ADD_PARTIAL_SEGMENT,
                         {
                             "message": AgentServerMessageType.ADD_PARTIAL_SEGMENT.value,
-                            "segments": [s.to_dict(self._config.include_results) for s in interim_segments],
-                            "metadata": metadata,
+                            "segments": [s.model_dump(self._config.include_results) for s in interim_segments],
+                            "metadata": partial_metadata,
                         },
                     )
 
@@ -884,7 +960,7 @@ class VoiceAgentClient(AsyncClient):
     def _vad_evaluation(self, fragments: list[SpeechFragment]) -> None:
         """Emit a VAD event.
 
-        This will emit `SPEECH_STARTED` and `SPEECH_ENDED` events to the client and is
+        This will emit `SPEAKER_STARTED` and `SPEAKER_ENDED` events to the client and is
         based on valid transcription for active speakers. Ignored or speakers not in
         focus will not be considered an active participant.
 
@@ -929,14 +1005,14 @@ class VoiceAgentClient(AsyncClient):
                     AgentServerMessageType.SPEAKER_ENDED,
                     {
                         "message": AgentServerMessageType.SPEAKER_ENDED.value,
-                        "status": SpeakerVADStatus(speaker_id=current_speaker, is_active=False).to_dict(),
+                        "status": SpeakerVADStatus(speaker_id=current_speaker, is_active=False).model_dump(),
                     },
                 )
                 self.emit(
                     AgentServerMessageType.SPEAKER_STARTED,
                     {
                         "message": AgentServerMessageType.SPEAKER_STARTED.value,
-                        "status": SpeakerVADStatus(speaker_id=speaker, is_active=True).to_dict(),
+                        "status": SpeakerVADStatus(speaker_id=speaker, is_active=True).model_dump(),
                     },
                 )
 
@@ -958,7 +1034,7 @@ class VoiceAgentClient(AsyncClient):
             msg,
             {
                 "message": msg.value,
-                "status": SpeakerVADStatus(speaker_id=speaker, is_active=self._is_speaking).to_dict(),
+                "status": SpeakerVADStatus(speaker_id=speaker, is_active=self._is_speaking).model_dump(),
             },
         )
 
