@@ -5,12 +5,13 @@
 from __future__ import annotations
 
 import datetime
-from dataclasses import dataclass
-from dataclasses import field
 from enum import Enum
 from typing import Any
 from typing import Optional
 
+from pydantic import BaseModel
+from pydantic import Field
+
 from speechmatics.rt import AudioEncoding
 from speechmatics.rt import OperatingPoint
 
@@ -23,6 +24,16 @@ class EndOfUtteranceMode(str, Enum):
     ADAPTIVE = "adaptive"
 
 
+class TranscriptionUpdatePreset(str, Enum):
+    """Filter options for when to emit changes to transcription."""
+
+    COMPLETE = "complete"
+    COMPLETE_PLUS_TIMING = "complete_plus_timing"
+    WORDS = "words"
+    WORDS_PLUS_TIMING = "words_plus_timing"
+    TIMING = "timing"
+
+
 class DiarizationFocusMode(str, Enum):
     """Speaker focus mode for diarization."""
 
@@ -30,8 +41,7 @@ class DiarizationFocusMode(str, Enum):
     IGNORE = "ignore"
 
 
-@dataclass
-class AdditionalVocabEntry:
+class AdditionalVocabEntry(BaseModel):
     """Additional vocabulary entry.
 
     Parameters:
@@ -40,11 +50,10 @@ class AdditionalVocabEntry:
     """
 
     content: str
-    sounds_like: list[str] = field(default_factory=list)
+    sounds_like: list[str] = Field(default_factory=list)
 
 
-@dataclass
-class DiarizationKnownSpeaker:
+class DiarizationKnownSpeaker(BaseModel):
     """Known speakers for speaker diarization.
 
     Parameters:
@@ -56,8 +65,7 @@ class DiarizationKnownSpeaker:
     speaker_identifiers: list[str]
 
 
-@dataclass
-class DiarizationSpeakerConfig:
+class DiarizationSpeakerConfig(BaseModel):
     """Speaker Diarization Config.
 
     List of speakers to focus on, ignore and how to deal with speakers that are not
@@ -86,13 +94,12 @@ class DiarizationSpeakerConfig:
             Defaults to `DiarizationFocusMode.RETAIN`.
     """
 
-    focus_speakers: list[str] = field(default_factory=list)
-    ignore_speakers: list[str] = field(default_factory=list)
+    focus_speakers: list[str] = Field(default_factory=list)
+    ignore_speakers: list[str] = Field(default_factory=list)
     focus_mode: DiarizationFocusMode = DiarizationFocusMode.RETAIN
 
 
-@dataclass
-class VoiceAgentConfig:
+class VoiceAgentConfig(BaseModel):
     """Voice Agent configuration.
 
     A framework-independent configuration object for the Speechmatics Voice Agent. This uses
@@ -101,28 +108,28 @@ class VoiceAgentConfig:
 
     Parameters:
         operating_point: Operating point for transcription accuracy vs. latency tradeoff. It is
-            recommended to use OperatingPoint.ENHANCED for most use cases. Defaults to
-            OperatingPoint.ENHANCED.
+            recommended to use `OperatingPoint.ENHANCED` for most use cases. Defaults to
+            `OperatingPoint.ENHANCED`.
 
-        domain: Domain for Speechmatics API. Defaults to None.
+        domain: Domain for Speechmatics API. Defaults to `None`.
 
         language: Language code for transcription. Defaults to `en`.
 
-        output_locale: Output locale for transcription, e.g. `en-GB`. Defaults to None.
+        output_locale: Output locale for transcription, e.g. `en-GB`. Defaults to `None`.
 
         max_delay: Maximum delay in seconds for transcription. This forces the STT engine to
             speed up the processing of transcribed words and reduces the interval between partial
-            and final results. Lower values can have an impact on accuracy. Defaults to 0.7.
+            and final results. Lower values can have an impact on accuracy. Defaults to `0.7`.
 
         end_of_utterance_silence_trigger: Maximum delay in seconds for end of utterance trigger.
             The delay is used to wait for any further transcribed words before emitting the final
             word frames. The value must be lower than max_delay.
-            Defaults to 0.2.
+            Defaults to `0.2`.
 
         end_of_utterance_max_delay: Maximum delay in seconds for end of utterance delay.
             The delay is used to wait for any further transcribed words before emitting the final
             word frames. The value must be greater than end_of_utterance_silence_trigger.
-            Defaults to 10.0.
+            Defaults to `10.0`.
 
         end_of_utterance_mode: End of utterance delay mode. When ADAPTIVE is used, the delay
             can be adjusted on the content of what the most recent speaker has said, such as
@@ -139,21 +146,21 @@ class VoiceAgentConfig:
         punctuation_overrides: Punctuation overrides. This allows you to override the punctuation
             in the STT engine. This is useful for languages that use different punctuation
             than English. See documentation for more information.
-            Defaults to None.
+            Defaults to `None`.
 
         enable_diarization: Enable speaker diarization. When enabled, the STT engine will
             determine and attribute words to unique speakers. The speaker_sensitivity
             parameter can be used to adjust the sensitivity of diarization.
-            Defaults to False.
+            Defaults to `False`.
 
         speaker_sensitivity: Diarization sensitivity. A higher value increases the sensitivity
             of diarization and helps when two or more speakers have similar voices.
-            Defaults to 0.5.
+            Defaults to `0.5`.
 
         max_speakers: Maximum number of speakers to detect. This forces the STT engine to cluster
             words into a fixed number of speakers. It should not be used to limit the number of
             speakers, unless it is clear that there will only be a known number of speakers.
-            Defaults to None.
+            Defaults to `None`.
 
         prefer_current_speaker: Prefer current speaker ID. When set to true, groups of words close
             together are given extra weight to be identified as the same speaker.
@@ -177,8 +184,18 @@ class VoiceAgentConfig:
         enable_preview_features: Enable preview features using a preview endpoint provided by
             Speechmatics. Defaults to False.
 
-        sample_rate: Audio sample rate for streaming. Defaults to 16000.
-        audio_encoding: Audio encoding format. Defaults to AudioEncoding.PCM_S16LE.
+        transcription_update_preset: Emit segments when the text content or word timings change.
+            Options are: `COMPLETE` (emit on changes to text content), `COMPLETE_PLUS_TIMING`
+            (emit on changes to text content and word timings), `WORDS` (emit on changes to word
+            content, without punctuation), `WORDS_PLUS_TIMING` (emit on changes to word content
+            and word timings), and `TIMING` (emit on changes to word timings, not recommended).
+            Defaults to `TranscriptionUpdatePreset.COMPLETE`.
+
+        audio_buffer_length: Length of audio buffer to extract slices of recent audio for post-processing
+            by end of thought models. Defaults to 0.0 seconds.
+
+        sample_rate: Audio sample rate for streaming. Defaults to `16000`.
+        audio_encoding: Audio encoding format. Defaults to `AudioEncoding.PCM_S16LE`.
     """
 
     # Service configuration
@@ -192,7 +209,7 @@ class VoiceAgentConfig:
     end_of_utterance_silence_trigger: float = 0.2
     end_of_utterance_max_delay: float = 10.0
     end_of_utterance_mode: EndOfUtteranceMode = EndOfUtteranceMode.FIXED
-    additional_vocab: list[AdditionalVocabEntry] = field(default_factory=list)
+    additional_vocab: list[AdditionalVocabEntry] = Field(default_factory=list)
     punctuation_overrides: Optional[dict] = None
 
     # Diarization
@@ -200,12 +217,14 @@ class VoiceAgentConfig:
     speaker_sensitivity: float = 0.5
     max_speakers: Optional[int] = None
     prefer_current_speaker: bool = False
-    speaker_config: DiarizationSpeakerConfig = field(default_factory=DiarizationSpeakerConfig)
-    known_speakers: list[DiarizationKnownSpeaker] = field(default_factory=list)
+    speaker_config: DiarizationSpeakerConfig = Field(default_factory=DiarizationSpeakerConfig)
+    known_speakers: list[DiarizationKnownSpeaker] = Field(default_factory=list)
 
     # Advanced features
     include_results: bool = False
     enable_preview_features: bool = False
+    transcription_update_preset: TranscriptionUpdatePreset = TranscriptionUpdatePreset.COMPLETE
+    audio_buffer_length: float = 0.0
 
     # Audio
     sample_rate: int = 16000
@@ -330,6 +349,7 @@ class AnnotationFlags(str, Enum):
     UPDATED_FINALS = "updated_finals"
     UPDATED_PARTIALS = "updated_partials"
     UPDATED_SPEAKERS = "updated_speakers"
+    UPDATED_WORD_TIMINGS = "updated_word_timings"
     FINALIZED = "finalized"
 
     # Content of segments
@@ -359,8 +379,7 @@ class AnnotationFlags(str, Enum):
     END_OF_UTTERANCE = "end_of_utterance"
 
 
-@dataclass
-class LanguagePackInfo:
+class LanguagePackInfo(BaseModel):
     """Information about the language pack used in a session.
 
     Attributes:
@@ -371,35 +390,14 @@ class LanguagePackInfo:
         writing_direction (str): The writing direction ('ltr' or 'rtl').
     """
 
-    adapted: bool
-    itn: bool
-    language_description: str
-    word_delimiter: str
-    writing_direction: str
-
-    @classmethod
-    def from_dict(cls, data: dict[str, Any]) -> LanguagePackInfo:
-        """Create LanguagePackInfo from dictionary."""
-        return cls(
-            adapted=data.get("adapted", False),
-            itn=data.get("itn", True),
-            language_description=data.get("language_description", "English"),
-            word_delimiter=data.get("word_delimiter", " "),
-            writing_direction="rtl" if data.get("writing_direction", "left-to-right") == "right-to-left" else "ltr",
-        )
-
-    def to_dict(self) -> dict[str, Any]:
-        return {
-            "adapted": self.adapted,
-            "itn": self.itn,
-            "language_description": self.language_description,
-            "word_delimiter": self.word_delimiter,
-            "writing_direction": self.writing_direction,
-        }
+    adapted: bool = False
+    itn: bool = True
+    language_description: str = "English"
+    word_delimiter: str = " "
+    writing_direction: str = "ltr"
 
 
-@dataclass
-class SessionInfo:
+class ClientSessionInfo(BaseModel):
     """Information about the session.
 
     Attributes:
@@ -414,14 +412,6 @@ class SessionInfo:
     base_time: datetime.datetime
     language_pack_info: LanguagePackInfo
 
-    def to_dict(self) -> dict[str, Any]:
-        return {
-            "config": self.config,
-            "session_id": self.session_id,
-            "base_time": self.base_time,
-            "language_pack_info": self.language_pack_info.to_dict(),
-        }
-
 
 class AnnotationResult(list):
     """Processing result."""
@@ -459,13 +449,8 @@ class AnnotationResult(list):
             return set(self) == set(other)
         return False
 
-    # def __repr__(self) -> str:
-    #     """Return a representation of the object."""
-    #     return repr([f.value for f in self])
 
-
-@dataclass
-class SpeechFragment:
+class SpeechFragment(BaseModel):
     """Fragment of a speech event.
 
     As the transcript is processed (partials and finals), a list of SpeechFragments
@@ -507,9 +492,10 @@ class SpeechFragment:
     result: Optional[Any] = None
     annotation: Optional[AnnotationResult] = None
 
+    model_config = {"arbitrary_types_allowed": True}
 
-@dataclass
-class SpeakerSegment:
+
+class SpeakerSegment(BaseModel):
     """SpeechFragment items grouped by speaker_id and whether the speaker is active.
 
     Parameters:
@@ -526,9 +512,11 @@ class SpeakerSegment:
     is_active: bool = False
     timestamp: Optional[str] = None
     language: Optional[str] = None
-    fragments: list[SpeechFragment] = field(default_factory=list)
-    text: str | None = None
-    annotation: AnnotationResult = field(default_factory=AnnotationResult)
+    fragments: list[SpeechFragment] = Field(default_factory=list)
+    text: Optional[str] = None
+    annotation: AnnotationResult = Field(default_factory=AnnotationResult)
+
+    model_config = {"arbitrary_types_allowed": True}
 
     @property
     def start_time(self) -> float:
@@ -540,57 +528,57 @@ class SpeakerSegment:
         """Return the end time of the segment."""
         return self.fragments[-1].end_time if self.fragments else 0.0
 
-    def to_dict(self, include_results: bool = False) -> dict[str, Any]:
-        """Return a dictionary representation of the object."""
-        return {
-            "speaker_id": self.speaker_id,
-            "is_active": self.is_active,
-            "timestamp": self.timestamp,
-            "language": self.language,
-            "text": self.text,
-            "annotation": self.annotation,
-            **({"results": [f.result for f in self.fragments]} if include_results else {}),
-        }
+    def model_dump(self, include_results: bool = False, **kwargs: Any) -> dict[str, Any]:
+        """Override model_dump to control fragments/results inclusion."""
 
-    def __str__(self) -> str:
-        """Return a string representation of the object."""
-        meta = self.to_dict(include_results=True)
-        return f"SpeakerSegment({', '.join(f'{k}={v}' for k, v in meta.items())})"
+        # Always exclude fragments from the base dump
+        kwargs["exclude"] = {"fragments"}
+        data: dict[str, Any] = super().model_dump(**kwargs)
+
+        # Add timing information
+        data["start_time"] = self.start_time
+        data["end_time"] = self.end_time
+
+        # Add results if requested
+        if include_results:
+            data["results"] = [f.result for f in self.fragments]
+
+        # Return the dump
+        return data
 
 
-@dataclass
-class SpeakerSegmentView:
+class SpeakerSegmentView(BaseModel):
     """View for speaker fragments.
 
     Parameters:
-        session: SessionInfo object.
+        session: ClientSessionInfo object.
         fragments: List of fragments.
         focus_speakers: List of speakers to focus on or None.
     """
 
-    session: SessionInfo
+    session: ClientSessionInfo
     fragments: list[SpeechFragment]
-    segments: list[SpeakerSegment]
+    segments: list[SpeakerSegment] = Field(default_factory=list)
+    focus_speakers: Optional[list[str]] = None
 
     def __init__(
         self,
-        session: SessionInfo,
+        session: ClientSessionInfo,
         fragments: list[SpeechFragment],
         focus_speakers: Optional[list[str]] = None,
         annotate_segments: bool = True,
-    ):
-        self.session = session
-        self.fragments = fragments
-        self.focus_speakers = focus_speakers
-
+        **data: Any,
+    ) -> None:
         # Process fragments into a list of segments
-        self.segments = FragmentUtils.segment_list_from_fragments(
+        segments = FragmentUtils.segment_list_from_fragments(
             session=session,
             fragments=fragments,
             focus_speakers=focus_speakers,
             annotate_segments=annotate_segments,
         )
 
+        super().__init__(session=session, fragments=fragments, segments=segments, focus_speakers=focus_speakers, **data)
+
     @property
     def start_time(self) -> float:
         return self.fragments[0].start_time if self.fragments else 0.0
@@ -666,8 +654,7 @@ class SpeakerSegmentView:
         )
 
 
-@dataclass
-class SpeakerVADStatus:
+class SpeakerVADStatus(BaseModel):
     """Emitted when a speaker starts or ends speaking.
 
     The speaker id is taken from the last word in the segment when
@@ -681,24 +668,18 @@ class SpeakerVADStatus:
     is_active: bool
     speaker_id: Optional[str] = None
 
-    def to_dict(self) -> dict[str, Any]:
-        return {
-            "is_active": self.is_active,
-            "speaker_id": self.speaker_id,
-        }
-
 
 class FragmentUtils:
     """Set of utility functions for working with SpeechFragment and SpeakerSegment objects."""
 
     @staticmethod
     def format_segment_text(
-        session: SessionInfo, segment: SpeakerSegment, format: str = "{text}", words_only: bool = False
+        session: ClientSessionInfo, segment: SpeakerSegment, format: str = "{text}", words_only: bool = False
     ) -> str:
         """Format a segment's text based on the language pack info.
 
         Args:
-            session: SessionInfo object.
+            session: ClientSessionInfo object.
             segment: SpeakerSegment object.
             words_only: Whether to include only word fragments.
 
@@ -715,12 +696,16 @@ class FragmentUtils:
         else:
             fragments = segment.fragments
 
-        # Assemble the text (LTR)
+        # Assemble the text
+        previous_frag: Optional[SpeechFragment] = None
         for frag in fragments:
-            if content == "" or frag.attaches_to == "previous":
+            if not previous_frag:
+                content = frag.content
+            elif frag.attaches_to == "previous" or previous_frag.attaches_to == "next":
                 content += frag.content
             else:
                 content += session.language_pack_info.word_delimiter + frag.content
+            previous_frag = frag
 
         # Return the formatted text
         return format.format(
@@ -729,12 +714,14 @@ class FragmentUtils:
                 "text": content,
                 "ts": segment.timestamp,
                 "lang": segment.language,
+                "start_time": fragments[0].start_time,
+                "end_time": fragments[-1].end_time,
             }
         )
 
     @staticmethod
     def segment_list_from_fragments(
-        session: SessionInfo,
+        session: ClientSessionInfo,
         fragments: list[SpeechFragment],
         focus_speakers: Optional[list[str]] = None,
         annotate_segments: bool = True,
@@ -742,7 +729,7 @@ class FragmentUtils:
         """Create SpeakerSegment objects from a list of SpeechFragment objects.
 
         Args:
-            session: SessionInfo object.
+            session: ClientSessionInfo object.
             fragments: List of SpeechFragment objects.
             focus_speakers: List of speakers to focus on or None.
             annotate_segments: Whether to annotate segments.
@@ -781,7 +768,7 @@ class FragmentUtils:
 
     @staticmethod
     def segment_from_fragments(
-        session: SessionInfo,
+        session: ClientSessionInfo,
         fragments: list[SpeechFragment],
         focus_speakers: Optional[list[str]] = None,
         annotate: bool = True,
@@ -796,7 +783,7 @@ class FragmentUtils:
         be removed.
 
         Args:
-            session: SessionInfo object.
+            session: ClientSessionInfo object.
             fragments: List of SpeechFragment objects.
             focus_speakers: List of speakers to focus on.
             annotate: Whether to annotate the segment.
@@ -925,14 +912,14 @@ class FragmentUtils:
 
     @staticmethod
     def compare_views(
-        session: SessionInfo, view1: SpeakerSegmentView, view2: Optional[SpeakerSegmentView]
+        session: ClientSessionInfo, view1: SpeakerSegmentView, view2: Optional[SpeakerSegmentView]
     ) -> AnnotationResult:
         """Compare two SpeakerSegmentView objects and return the differences.
 
         View 1 (new) is compared to view 2 (old).
 
         Args:
-            session: SessionInfo object.
+            session: ClientSessionInfo object.
             view1: The first SpeakerSegmentView object to compare.
             view2: The second SpeakerSegmentView object to compare to or None.
 
@@ -960,6 +947,12 @@ class FragmentUtils:
             if view1_stripped_str.lower() != view2_stripped_str.lower():
                 result.add(AnnotationFlags.UPDATED_STRIPPED_LCASE)
 
+            # Word timings
+            view1_timings_str: str = view1.format_view_text(format="|{start_time}-{end_time}|", words_only=True)
+            view2_timings_str: str = view2.format_view_text(format="|{start_time}-{end_time}|", words_only=True)
+            if view1_timings_str != view2_timings_str:
+                result.add(AnnotationFlags.UPDATED_WORD_TIMINGS)
+
             # Partials, finals and speakers
             if view1.final_count != view2.final_count:
                 result.add(AnnotationFlags.UPDATED_FINALS)
